use meta_compile::{
    compile_bind, compile_dec, compile_entry, compile_fun, compile_import, compile_module,
};

use crate::parser::{Ast, Localised, Location};

pub mod backend;
mod meta_compile;

use std::fmt;

#[cfg(test)]
mod tests;

#[derive(Debug, PartialEq, Default, Clone)]
pub enum Capability {
    #[default]
    Shader,
}

#[derive(Debug, PartialEq, Default, Clone)]
pub enum ExecutionMode {
    #[default]
    OriginUpperLeft,
}

#[derive(Debug, PartialEq, Default, Clone)]
pub enum ExecutionModel {
    #[default]
    Fragment,
    Vertex,
}

#[derive(Debug, PartialEq, Default, Clone)]
pub struct EntryPoint {
    pub execution_model: ExecutionModel,
    pub execution_mode: ExecutionMode,
    pub name: String,
    pub interface: Vec<String>,
}

#[derive(Debug, PartialEq, Default, Clone)]
pub enum AddressingModel {
    #[default]
    Logical,
    Physical32,
    Physical64,
    PhysicalStorageBuffer64,
}
#[derive(Debug, PartialEq, Default, Clone)]
pub enum MemoryModel {
    #[default]
    GLSL450,
    OpenCL,
    VulkanKHR,
    Simple,
}

#[derive(Debug, PartialEq, Default, Clone)]
pub struct Memory {
    pub addressing_model: AddressingModel,
    pub memory_model: MemoryModel,
}

#[derive(Debug, PartialEq, Default, Clone)]
pub enum BuiltinDecoration {
    #[default]
    FragCoord,
}

#[derive(Debug, PartialEq, Clone)]
pub enum Decoration {
    BuiltIn(BuiltinDecoration),
    Location(u32),
}

#[derive(Debug, PartialEq, Default, Clone)]
pub enum StorageClass {
    #[default]
    Undefined,
    Input,
    Output,
}

impl fmt::Display for StorageClass {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match self {
            StorageClass::Undefined => write!(f, "Undefined"),
            StorageClass::Input => write!(f, "Input"),
            StorageClass::Output => write!(f, "Output"),
        }
    }
}

#[derive(Debug, PartialEq, Default, Clone)]
pub struct GlobalVariable {
    pub name: String,
    pub typ: String,
    pub storage_class: StorageClass,
    pub decorations: Vec<Decoration>,
}

#[derive(Debug, PartialEq, Default, Clone)]
pub struct Instruction {
    pub result_id: Option<String>,
    pub op: String,
    pub operands: Vec<String>,
}

#[derive(Debug, PartialEq, Default, Clone)]
pub struct Function {
    pub name: String,
    pub return_type: String,
    pub arguments: Vec<String>,
    pub body: Option<Vec<Instruction>>,
    pub ast: Option<Ast>,
}

#[derive(Debug, PartialEq, Default, Clone)]
pub struct Import {
    pub name: String,
    pub value: String,
}

#[derive(Debug, PartialEq, Default, Clone)]
pub struct Module {
    pub capabilities: Vec<Capability>,
    pub entry_points: Vec<EntryPoint>,
    pub globals: Vec<GlobalVariable>,
    pub functions: Vec<Function>,
    pub memory_model: Memory,
    pub imports: Vec<Import>,
}

#[derive(Debug)]
pub enum CompilerError {
    ExpectedSymbol(Location),
    ExpectedList(Location),
    UnknownKeyword(String, Location),
    ExpectedOneOf(Vec<String>, String, Location),
    TrailingTokens(Location),
    UnknownBind(String, Location),
    ExpectedType(Location),
}

fn expect_symbol(ast: Option<Ast>, loc: &Location) -> Result<Localised<String>, CompilerError> {
    match ast {
        Some(Ast::Symbol(s)) => Ok(s),
        Some(l) => Err(CompilerError::ExpectedSymbol(l.location().clone())),
        None => Err(CompilerError::ExpectedSymbol(loc.tail())),
    }
}

fn expect_list(ast: Option<Ast>, loc: &Location) -> Result<Localised<Vec<Ast>>, CompilerError> {
    match ast {
        Some(Ast::List(l)) => Ok(l),
        Some(l) => Err(CompilerError::ExpectedList(l.location().clone())),
        None => Err(CompilerError::ExpectedList(loc.tail())),
    }
}

fn expect_one_of<T>(
    options: &[T],
    Localised { location, item }: Localised<String>,
) -> Result<Localised<String>, CompilerError>
where
    T: PartialEq<String> + Into<String> + Clone,
{
    if options.iter().find(|e| **e == item).is_some() {
        Ok(Localised { location, item })
    } else {
        Err(CompilerError::ExpectedOneOf(
            options
                .iter()
                .map(|t| Into::<String>::into(t.clone()))
                .collect(),
            item,
            location,
        ))
    }
}

fn expect_empty_ast<I: Iterator<Item = Ast>>(mut iter: I) -> Result<(), CompilerError> {
    match iter.next() {
        Some(a) => Err(CompilerError::TrailingTokens(a.location().clone())),
        None => Ok(()),
    }
}

fn expect_empty<T, I: Iterator<Item = T>>(
    mut iter: I,
    location: Location,
) -> Result<(), CompilerError> {
    match iter.next() {
        Some(_) => Err(CompilerError::TrailingTokens(location)),
        None => Ok(()),
    }
}

pub fn meta_compile(ast: Ast) -> Result<Module, CompilerError> {
    let mut module: Module = Module::default();
    let Ast::Root(root) = ast else {
        panic!("Non-root ast"); // compiler error, fine to panic
    };
    for node in root {
        let Localised {
            location,
            item: list,
        } = expect_list(Some(node), &Location::All)?;
        let mut iter = list.into_iter();

        match iter.next() {
            Some(Ast::Symbol(sym)) => match sym.as_str() {
                "module" => {
                    compile_module(&mut module, iter, location)?;
                }
                "import" => {
                    compile_import(&mut module, iter, location)?;
                }
                "bind" => {
                    compile_bind(&mut module, iter, location)?;
                }
                "dec" => {
                    compile_dec(&mut module, iter, location)?;
                }
                "entry" => {
                    compile_entry(&mut module, iter, location)?;
                }
                "fun" => {
                    compile_fun(&mut module, iter, location)?;
                }
                _ => {
                    let Localised { location, item } = sym;
                    return Err(CompilerError::UnknownKeyword(item, location));
                }
            },
            Some(l) => return Err(CompilerError::ExpectedSymbol(l.location().clone())),
            None => return Err(CompilerError::ExpectedSymbol(location)),
        }
    }
    Ok(module)
}
